@@ -3,7 +3,7 @@
 use std::ops::ControlFlow;
 
 use either::Either;
-use hir::{AsAssocItem, HasAttrs, HasVisibility, Semantics, sym};
+use hir::{AsAssocItem, HasAttrs, HasVisibility, Semantics};
 use ide_db::{
     FxHashMap, RootDatabase, SymbolKind,
     defs::{Definition, IdentClass, NameClass, NameRefClass},
@@ -399,7 +399,7 @@ fn highlight_name_ref(
             highlight_def(sema, krate, field_ref.into(), edition, true)
         }
         NameRefClass::ExternCrateShorthand { decl, krate: resolved_krate } => {
-            let mut h = HlTag::Symbol(SymbolKind::Module).into();
+            let mut h = HlTag::Symbol(SymbolKind::CrateRoot).into();
 
             if krate.as_ref().is_some_and(|krate| resolved_krate != *krate) {
                 h |= HlMod::Library;
@@ -413,7 +413,10 @@ fn highlight_name_ref(
             if is_from_builtin_crate {
                 h |= HlMod::DefaultLibrary;
             }
-            h |= HlMod::CrateRoot;
+            let is_deprecated = resolved_krate.attrs(sema.db).is_deprecated();
+            if is_deprecated {
+                h |= HlMod::Deprecated;
+            }
             h
         }
     };
@@ -491,16 +494,15 @@ pub(super) fn highlight_def(
             (Highlight::new(HlTag::Symbol(SymbolKind::Field)), Some(field.attrs(sema.db)))
         }
         Definition::TupleField(_) => (Highlight::new(HlTag::Symbol(SymbolKind::Field)), None),
-        Definition::Crate(krate) => (
-            Highlight::new(HlTag::Symbol(SymbolKind::Module)) | HlMod::CrateRoot,
-            Some(krate.attrs(sema.db)),
-        ),
+        Definition::Crate(krate) => {
+            (Highlight::new(HlTag::Symbol(SymbolKind::CrateRoot)), Some(krate.attrs(sema.db)))
+        }
         Definition::Module(module) => {
-            let mut h = Highlight::new(HlTag::Symbol(SymbolKind::Module));
-            if module.is_crate_root() {
-                h |= HlMod::CrateRoot;
-            }
-
+            let h = Highlight::new(HlTag::Symbol(if module.is_crate_root(db) {
+                SymbolKind::CrateRoot
+            } else {
+                SymbolKind::Module
+            }));
             (h, Some(module.attrs(sema.db)))
         }
         Definition::Function(func) => {
@@ -658,8 +660,7 @@ pub(super) fn highlight_def(
             (h, None)
         }
         Definition::ExternCrateDecl(extern_crate) => {
-            let mut highlight =
-                Highlight::new(HlTag::Symbol(SymbolKind::Module)) | HlMod::CrateRoot;
+            let mut highlight = Highlight::new(HlTag::Symbol(SymbolKind::CrateRoot));
             if extern_crate.alias(db).is_none() {
                 highlight |= HlMod::Library;
             }
@@ -697,7 +698,7 @@ pub(super) fn highlight_def(
     }
 
     if let Some(attrs) = attrs
-        && attrs.by_key(sym::deprecated).exists()
+        && attrs.is_deprecated()
     {
         h |= HlMod::Deprecated;
     }
@@ -743,11 +744,11 @@ fn highlight_method_call(
         h |= HlMod::Trait;
     }
 
-    let def_crate = func.module(sema.db).krate();
+    let def_crate = func.module(sema.db).krate(sema.db);
     let is_from_other_crate = krate.as_ref().map_or(false, |krate| def_crate != *krate);
     let is_from_builtin_crate = def_crate.is_builtin(sema.db);
     let is_public = func.visibility(sema.db) == hir::Visibility::Public;
-    let is_deprecated = func.attrs(sema.db).by_key(sym::deprecated).exists();
+    let is_deprecated = func.attrs(sema.db).is_deprecated();
 
     if is_from_other_crate {
         h |= HlMod::Library;
@@ -801,6 +802,7 @@ fn highlight_name_by_syntax(name: ast::Name) -> Highlight {
         TYPE_PARAM => SymbolKind::TypeParam,
         RECORD_FIELD => SymbolKind::Field,
         MODULE => SymbolKind::Module,
+        EXTERN_CRATE => SymbolKind::CrateRoot,
         FN => SymbolKind::Function,
         CONST => SymbolKind::Const,
         STATIC => SymbolKind::Static,
@@ -831,7 +833,7 @@ fn highlight_name_ref_by_syntax(
     };
 
     match parent.kind() {
-        EXTERN_CRATE => HlTag::Symbol(SymbolKind::Module) | HlMod::CrateRoot,
+        EXTERN_CRATE => HlTag::Symbol(SymbolKind::CrateRoot).into(),
         METHOD_CALL_EXPR => ast::MethodCallExpr::cast(parent)
             .and_then(|it| highlight_method_call(sema, krate, &it, is_unsafe_node))
             .unwrap_or_else(|| SymbolKind::Method.into()),
